from .log_events import (
    log_event,
    error_log,
    warning_log,
    info_log,
    debug_log,
    debug_log_if
)
from .extras import (
    run_command,
    install_import
)

__all__ = [
    'log_event', 'error_log', 'warning_log',
    'info_log', 'debug_log', 'debug_log_if',
    'run_command', 'install_import'
]



#!python3

"""
Lumeo Bulk Deployer

Lumeo Bulk Deployer uploads & queues media file deployments, or creates live stream deployments on a specific gateway. 
Learn more at https://docs.lumeo.com/docs/bulk-deployer

Arguments:
    Authentication Args:
        --app_id APP_ID       Application (aka Workspace) ID
        --token TOKEN         Access (aka API) Token.

    Source Files (one of pattern, file_list, csv_file, s3_bucket or tag is required):
        --pattern PATTERN     Glob pattern for files to upload
        --file_list FILE_LIST
                                Comma separated list of file URIs to queue
        --csv_file CSV_FILE   CSV file containing file_uri and corresponding camera_external_id or camera_id
        --s3_bucket S3_BUCKET
                                S3 bucket name to use as source for files
        --s3_access_key_id S3_ACCESS_KEY_ID
                                S3 Access key ID
        --s3_secret_access_key S3_SECRET_ACCESS_KEY
                                S3 secret access key
        --s3_region S3_REGION
                                S3 region if using AWS S3 bucket. Either s3_region or s3_endpoint_url must be specified.
        --s3_endpoint_url S3_ENDPOINT_URL
                                S3 endpoint URL. Either s3_region or s3_endpoint_url must be specified.
        --s3_prefix S3_PREFIX
                                S3 path prefix to filter files. Optional.
        --tag TAG             Tag to apply to uploaded files. Can be tag uuid, tag name or tag path (e.g. "tag1/tag2/tag3").If specified without pattern/file_list/csv_file/s3_bucket, will process
                                existing files with that tag.

    Associate Files with Virtual Camera (gets pipeline & deployment config from camera):
        --camera_id CAMERA_ID
                                Camera ID of an existing camera, to associate with the uploaded files
        --camera_external_id CAMERA_EXTERNAL_ID
                                Use your own unique camera id to find or create a virtual camera, and associate with the uploaded files

    Source Live Streams (gateway_id + one of stream_id, stream_urls, or stream_tag is required):
        --gateway_id GATEWAY_ID
                                Gateway ID to create (not queue) deployment for processing live streams.
        --stream_id STREAM_ID
                                Stream ID of an existing stream to create a deployment for processing live streams.
        --stream_urls STREAM_URLS
                                Comma separated list of RTSP or HLS stream URLs to create streams for processing.
        --stream_tag STREAM_TAG
                                Creates one deployment for each stream with the specified tag or applies tag to created streams when used with stream_urls. Can be tag uuid, tag name or tag path (e.g.
                                "tag1/tag2/tag3").

    Deployment Args (needed for Live Stream Deployments or when camera not specified for File deployments):
        --pipeline_id PIPELINE_ID
                                Pipeline ID to queue deployment for processing. Required if camera_id / camera_external_id not specified.
        --deployment_config DEPLOYMENT_CONFIG
                                String containing a Deployment config JSON object. Video source in the config will be overridden by source files specified in this script. Ignored if camera_id or
                                camera_external_id specified. Optional.

    General Args:
        --deployment_prefix DEPLOYMENT_PREFIX
                                Prefix to use for deployment name. Optional.
        --delete_processed DELETE_PROCESSED
                                Delete successfully processed files from the local folder after uploading
        --log_level LOG_LEVEL
                                Log level (DEBUG, INFO, WARNING, ERROR, CRITICAL)
        --batch_size BATCH_SIZE
                                Number of concurrent uploads to process at a time. Default 5.
        --queue_size          Print the current queue size
"""

import argparse
import asyncio
import boto3
import colorlog
import csv
import glob
import json
import logging
import time
import uuid
from pathlib import Path

from lumeo.api import LumeoApiClient
from botocore.client import Config
from .deployer import LumeoDeployer

handler = logging.StreamHandler()
handler.setFormatter(colorlog.ColoredFormatter(
    "%(log_color)s%(asctime)s %(levelname)-8s %(message)s",
    datefmt='%Y-%m-%d %H:%M:%S',
    log_colors={
        'DEBUG': 'cyan',
        'INFO': 'green',
        'WARNING': 'yellow',
        'ERROR': 'red,bg_white',
        'CRITICAL': 'red,bg_white',
    },
))

logger = colorlog.getLogger()
logger.addHandler(handler)
logger.setLevel(logging.INFO)


class BulkDeployer:
    def __init__(self):
        self._api_client = None
        self._queuing_lock = asyncio.Lock()
        self._start_time = time.time()
        self._parser = self._create_args_parser()
        self._args = self._parser.parse_args()

    async def run(self):
        tasks = []

        if not self._validate_args(self._args):
            self._parser.print_usage()
            return

        self._api_client = LumeoApiClient(self._args.app_id, self._args.token)

        if self._args.queue_size and not (self._args.pattern or self._args.file_list or self._args.csv_file or self._args.s3_bucket or self._args.tag):
            await self._print_queue_size(self._api_client)
            return

        if self._args.tag:
            # Create a tag if another file source is provided and the tag is not found.
            create_tag = (self._args.csv_file or self._args.file_list or self._args.pattern or self._args.s3_bucket)
            self._args.tag = await self._get_tag_id(self._args.tag, create_tag)
            
        if self._args.stream_tag:
            # Create a tag if another live stream source is provided and the tag is not found.
            create_tag = (self._args.stream_urls)
            self._args.stream_tag = await self._get_tag_id(self._args.stream_tag, create_tag)
            
        tasks = await self._get_tasks()

        results = await self._process_tasks(tasks)

        self._log_results(results)

        # Get the deployment queue for this app
        await self._print_queue_size(self._api_client)
        return

    def _create_args_parser(self):
        parser = argparse.ArgumentParser(description="""Lumeo Bulk Deployer uploads & queues media file deployments, \
                                                        or creates live stream deployments on a specific gateway. \
                                                        Learn more at https://docs.lumeo.com/docs/bulk-deployer """)

        required_group = parser.add_argument_group('Authentication Args')
        required_group.add_argument('--app_id', required=True, help='Application (aka Workspace) ID')
        required_group.add_argument('--token', required=True, help='Access (aka API) Token.')

        file_source_group = parser.add_argument_group('Source Files (one of pattern, file_list, csv_file, s3_bucket or tag is required)')
        file_source_group.add_argument('--pattern', help='Glob pattern for files to upload')
        file_source_group.add_argument('--file_list', help='Comma separated list of file URIs to queue')
        file_source_group.add_argument('--csv_file', help='CSV file containing file_uri and corresponding camera_external_id or camera_id')
        file_source_group.add_argument('--s3_bucket', help='S3 bucket name to use as source for files')
        file_source_group.add_argument('--s3_access_key_id', help='S3 Access key ID')
        file_source_group.add_argument('--s3_secret_access_key', help='S3 secret access key')
        file_source_group.add_argument('--s3_region', help='S3 region if using AWS S3 bucket. Either s3_region or s3_endpoint_url must be specified.')
        file_source_group.add_argument('--s3_endpoint_url', help='S3 endpoint URL. Either s3_region or s3_endpoint_url must be specified.')
        file_source_group.add_argument('--s3_prefix', help='S3 path prefix to filter files. Optional.')
        file_source_group.add_argument('--tag', help='Tag to apply to uploaded files. Can be tag uuid, tag name or tag path (e.g. "tag1/tag2/tag3").'
                                          'If specified without pattern/file_list/csv_file/s3_bucket, will process existing files with that tag.')

        camera_group = parser.add_argument_group('Associate Files with Virtual Camera (gets pipeline & deployment config from camera)')
        camera_group.add_argument('--camera_id', help='Camera ID of an existing camera, to associate with the uploaded files')
        camera_group.add_argument('--camera_external_id', help='Use your own unique camera id to find or create a virtual camera, and associate with the uploaded files')

        live_stream_group = parser.add_argument_group('Source Live Streams (gateway_id + one of stream_id, stream_urls, or stream_tag is required)')
        live_stream_group.add_argument('--gateway_id', help='Gateway ID to create (not queue) deployment for processing live streams.')
        live_stream_group.add_argument('--stream_id', help='Stream ID of an existing stream to create a deployment for processing live streams.')
        live_stream_group.add_argument('--stream_urls', help='Comma separated list of RTSP or HLS stream URLs to create streams for processing.')
        live_stream_group.add_argument('--stream_tag', help='Creates one deployment for each stream with the specified tag or applies tag to created streams '
                                                            'when used with stream_urls. '
                                                            'Can be tag uuid, tag name or tag path (e.g. "tag1/tag2/tag3").')

        pipeline_group = parser.add_argument_group('Deployment Args (needed for Live Stream Deployments or when camera not specified for File deployments)')
        pipeline_group.add_argument('--pipeline_id', help='Pipeline ID to queue deployment for processing. Required if camera_id / camera_external_id not specified.')
        pipeline_group.add_argument('--deployment_config', help='String containing a Deployment config JSON object. Video source in the config will be overridden by source files specified in this script. Ignored if camera_id or camera_external_id specified. Optional.')

        other_group = parser.add_argument_group('General Args')
        other_group.add_argument('--deployment_prefix', help='Prefix to use for deployment name. Optional.')
        other_group.add_argument('--delete_processed', help='Delete successfully processed files from the local folder after uploading')
        other_group.add_argument('--log_level', default='INFO', help='Log level (DEBUG, INFO, WARNING, ERROR, CRITICAL)')
        other_group.add_argument('--batch_size', type=int, default=5, help='Number of concurrent uploads to process at a time. Default 5.')
        other_group.add_argument('--queue_size', action='store_true', help='Print the current queue size')

        return parser

    def _validate_args(self, args):
        if len(vars(args)) <= 1:
            return False

        logger.setLevel(args.log_level.upper())

        if not args.queue_size:
            if not args.gateway_id: 
                if not any([args.pattern, args.file_list, args.csv_file, args.s3_bucket, args.tag]):
                    logging.error("Please provide either a tag for already uploaded files, glob pattern, a file list, a csv file, or an S3 bucket.")
                    return False
                elif not (args.csv_file or any([args.camera_id, args.camera_external_id, args.pipeline_id])):
                    logging.error("Please provide either a camera_id or camera_external_id or pipeline_id if source isnt a csv file.")
                    return False
                elif args.s3_bucket and not all([args.access_key_id, args.secret_access_key]):
                    logging.error("Please provide AWS credentials when using an S3 bucket")
                    return False
                elif args.s3_bucket and not any([args.region, args.endpoint_url]):
                    print("Please provide AWS S3 region OR endpoint URL when using an S3 bucket")
                    return False
            else:
                if not any([args.stream_id, args.stream_urls, args.stream_tag]):
                    logging.error("Gateway ID provided. Please also provide either a stream_id, stream_urls, or stream_tag.")
                    return False
                elif not args.pipeline_id:
                    logging.error("Gateway ID provided. Please also provide a pipeline_id to create deployments for live streams.")
                    return False

        if args.deployment_config:
            try:
                args.deployment_config = json.loads(args.deployment_config)
            except json.JSONDecodeError:
                logging.error("Invalid deployment config JSON provided")
                return False
        else:
            args.deployment_config = {}

        return True
    
    async def _get_tasks(self):
        if not self._args.gateway_id:
            if self._args.csv_file:
                return await self._process_csv_file(self._args, self._api_client, self._queuing_lock)
            elif self._args.file_list:
                return await self._process_file_list(self._args, self._api_client, self._queuing_lock)
            elif self._args.pattern:
                return await self._process_glob_pattern(self._args, self._api_client, self._queuing_lock)
            elif self._args.s3_bucket:
                return await self._process_s3_bucket(self._args, self._api_client, self._queuing_lock)
            elif self._args.tag:
                return await self._process_existing_files_with_tag(self._args, self._api_client, self._queuing_lock)
        else:
            if self._args.stream_id:
                return await self._process_existing_live_stream(self._args, self._api_client, self._queuing_lock)
            elif self._args.stream_urls:
                return await self._process_live_stream_list(self._args, self._api_client, self._queuing_lock)
            elif self._args.stream_tag:
                return await self._process_existing_live_streams_with_tag(self._args, self._api_client, self._queuing_lock)
        return []        
    
    async def _process_csv_file(self, args, api_client, queuing_lock):
        tasks = []
        with open(args.csv_file, 'r') as f:
            reader = csv.reader(f)
            next(reader)  # Skip the header
            for row in reader:
                file_uri = None
                camera_external_id = args.camera_external_id
                camera_id = args.camera_id
                pipeline_id = args.pipeline_id
                deployment_config = args.deployment_config
                file_uri = row[0].strip() if len(row[0].strip()) > 0 else None
                if file_uri and not file_uri.startswith('#'):
                    if len(row) > 1:
                        camera_external_id = row[1].strip() if len(row[1].strip()) > 0 else camera_external_id
                    if len(row) > 2:
                        camera_id = row[2].strip() if len(row[2].strip()) > 0 else camera_id
                    if len(row) > 3:
                        pipeline_id = row[3].strip() if len(row[3].strip()) > 0 else pipeline_id
                    if len(row) > 4:
                        deployment_config_str = row[4].strip() if len(row[4].strip()) > 0 else "{}"
                        try:
                            deployment_config_dict = json.loads(deployment_config_str)
                        except json.JSONDecodeError:
                            logging.error(f"Invalid deployment config JSON provided in row {reader.line_num} : {row}. Skipping.")
                            continue
                        deployment_config = deployment_config_dict

                    if not any([camera_external_id, camera_id, pipeline_id]):
                        logging.error(f"No camera_id, camera_external_id or pipeline_id provided in row {reader.line_num} : {row}. Skipping.")
                        continue

                    tasks.append(LumeoDeployer(api_client, queuing_lock, file_uri=file_uri, tag_id=args.tag,
                                               camera_external_id=camera_external_id, camera_id=camera_id,
                                               pipeline_id=pipeline_id, deployment_config=deployment_config,
                                               deployment_prefix=args.deployment_prefix, 
                                               delete_processed_files=args.delete_processed).process())

                else:
                    logging.error(f"Invalid file URI provided in CSV file row {reader.line_num}: {row}. Skipping.")
        return tasks

    async def _process_file_list(self, args, api_client, queuing_lock):
        tasks = []
        for file_uri in args.file_list.split(','):
            tasks.append(LumeoDeployer(api_client, queuing_lock, file_uri=file_uri, tag_id=args.tag,
                                       camera_external_id=args.camera_external_id, camera_id=args.camera_id,
                                       pipeline_id=args.pipeline_id, deployment_config=args.deployment_config,
                                       deployment_prefix=args.deployment_prefix, 
                                       delete_processed_files=args.delete_processed).process())
        return tasks

    async def _process_glob_pattern(self, args, api_client, queuing_lock):
        tasks = []
        for file_path in glob.glob(args.pattern):
            tasks.append(LumeoDeployer(api_client, queuing_lock, file_path=file_path, tag_id=args.tag,
                                        camera_external_id=args.camera_external_id, camera_id=args.camera_id,
                                        pipeline_id=args.pipeline_id, deployment_config=args.deployment_config,
                                        deployment_prefix=args.deployment_prefix, 
                                        delete_processed_files=args.delete_processed).process())
        return tasks

    async def _process_s3_bucket(self, args, api_client, queuing_lock):
        s3_file_list = await self._get_s3_file_list(args.bucket, args.access_key_id, args.secret_access_key, args.region, args.endpoint_url, args.prefix)
        tasks = []
        for signed_url in s3_file_list:
            tasks.append(LumeoDeployer(api_client, queuing_lock, file_uri=signed_url, tag_id=args.tag,
                                        camera_external_id=args.camera_external_id, camera_id=args.camera_id,
                                        pipeline_id=args.pipeline_id, deployment_config=args.deployment_config,
                                        deployment_prefix=args.deployment_prefix).process())
        return tasks

    async def _process_existing_files_with_tag(self, args, api_client, queuing_lock):
        tasks = []

        # Get all file streams with the specified tag ID
        logging.info(f"Getting existing file streams with tag {args.tag}")
        file_streams = await api_client.get_streams(args.tag, stream_type="file")
        for stream in file_streams:
            # Create a task for processing this file
            tasks.append(LumeoDeployer(api_client, queuing_lock, stream_id=stream['id'], 
                                        pipeline_id=args.pipeline_id,deployment_config=args.deployment_config,
                                        deployment_prefix=args.deployment_prefix).process())

        return tasks
        
    async def _process_existing_live_streams_with_tag(self, args, api_client, queuing_lock):
        tasks = []
        logging.info(f"Getting existing live streams with tag {args.stream_tag}")
        live_streams = await api_client.get_streams(args.stream_tag, stream_type="rtsp")
        for stream in live_streams:
            # Create a task for processing this file
            tasks.append(LumeoDeployer(api_client, queuing_lock, stream_id=stream['id'],
                                        gateway_id=args.gateway_id, pipeline_id=args.pipeline_id,
                                        deployment_config=args.deployment_config,
                                        deployment_prefix=args.deployment_prefix).process())
        return tasks
    
    async def _process_existing_live_stream(self, args, api_client, queuing_lock):
        tasks = []  
        tasks.append(LumeoDeployer(api_client, queuing_lock, stream_id=args.stream_id,
                                        gateway_id=args.gateway_id, pipeline_id=args.pipeline_id,
                                        deployment_config=args.deployment_config,
                                        deployment_prefix=args.deployment_prefix).process())
        return tasks
    
    async def _process_live_stream_list(self, args, api_client, queuing_lock):
        tasks = []  
        for stream_url in args.stream_urls.split(','):
            tasks.append(LumeoDeployer(api_client, queuing_lock, stream_uri=stream_url, tag_id=args.stream_tag,
                                        gateway_id=args.gateway_id, pipeline_id=args.pipeline_id,
                                        deployment_config=args.deployment_config,
                                        deployment_prefix=args.deployment_prefix).process())
        return tasks    
    
    async def _get_tag_id(self, tag_path, create=False):
        tag_id = None
        # tag is to be applied. get or create tag path.
        try:
            uuid.UUID(tag_path)
            tag_id = tag_path
        except ValueError:
            tag_id = await self._api_client.get_tag_id_by_path(tag_path)
            if not tag_id and create:
                logging.info(f"Tag path '{tag_path}' not found. Creating...")
                tag_id = await self._api_client.create_tag_path(tag_path)
            if not tag_id:
                logging.error(f"Tag path '{tag_path}' not found.")
        return tag_id

    async def _get_s3_file_list(self, bucket_name, access_key_id, secret_access_key, region, endpoint_url=None, prefix=None):
        file_list = []
        s3_config = {
            'aws_access_key_id': access_key_id,
            'aws_secret_access_key': secret_access_key,
            'config': Config(signature_version='s3v4')
        }

        if endpoint_url:
            s3_config['endpoint_url'] = endpoint_url
        else:
            s3_config['region_name'] = region

        s3 = boto3.client('s3', **s3_config)

        paginator = s3.get_paginator('list_objects_v2')
        pagination_params = {'Bucket': bucket_name}
        if prefix:
            pagination_params['Prefix'] = prefix

        for page in paginator.paginate(**pagination_params):
            for obj in page.get('Contents', []):
                signed_url = s3.generate_presigned_url('get_object',
                                                        Params={'Bucket': bucket_name,
                                                                'Key': obj['Key']},
                                                        ExpiresIn=604800)  # 1 week in seconds
                file_list.append(signed_url)

        return file_list

    async def _process_tasks(self, tasks):
        results = []
        sem = asyncio.Semaphore(self._args.batch_size)
        async def process_with_limit(task):
            async with sem:
                return await task

        tasks = [process_with_limit(task) for task in tasks]
        for completed in asyncio.as_completed(tasks):
            result = await completed
            results.append(result)
        return results

    def _log_results(self, results):
        end_time = time.time()
        successful_tasks = results.count(True)
        failed_tasks = results.count(False)    
        print(f"Finished queueing. Results : Total {len(results)}, Successful {successful_tasks}, Failed {failed_tasks}.")    
        print(f"Total processing time: {round(end_time - self._start_time, 2)} seconds")


    async def _print_queue_size(self, api_client: LumeoApiClient):
        # Get the deployment queue for this app
        deployment_queue_id = await api_client.get_deployment_queue_id()
        queue_entries = await api_client.get_queue_entries(deployment_queue_id)
        print(f"Current Queue size: {queue_entries['total_elements']}")

def run_main():
    asyncio.run(BulkDeployer().run())
        
# if __name__ == "__main__":
#     main()



###################################################################################
# Copyright (c) 2021 Rhombus Systems                                              #
#                                                                                 # 
# Permission is hereby granted, free of charge, to any person obtaining a copy    #
# of this software and associated documentation files (the "Software"), to deal   #
# in the Software without restriction, including without limitation the rights    #
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell       #
# copies of the Software, and to permit persons to whom the Software is           #
# furnished to do so, subject to the following conditions:                        #
#                                                                                 # 
# The above copyright notice and this permission notice shall be included in all  #
# copies or substantial portions of the Software.                                 #
#                                                                                 # 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR      #
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,        #
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE     #
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER          #
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,   #
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE   #
# SOFTWARE.                                                                       #
###################################################################################

import requests
import argparse
import sys
from . import rhombus_logging
from datetime import datetime, timedelta
import urllib3
import os
import subprocess

# just to prevent unnecessary logging since we are not verifying the host
from .rhombus_mpd_info import RhombusMPDInfo

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

_logger = rhombus_logging.get_logger("rhombus.CopyFootageToLocalStorage")

URI_FILE_ENDINGS = ["clip.mpd", "file.mpd"]


def get_segment_uri(mpd_uri, segment_name):
    for ending in URI_FILE_ENDINGS:
        if ending in mpd_uri:
            return mpd_uri.replace(ending, segment_name)

    return None


def get_segment_uri_index(rhombus_mpd_info, mpd_uri, index):
    segment_name = rhombus_mpd_info.segment_pattern.replace("$Number$", str(index + rhombus_mpd_info.start_index))
    return get_segment_uri(mpd_uri, segment_name)


class CopyFootageToLocalStorage:
    def __init__(self, cli_args):
        arg_parser = self.__initialize_argument_parser()
        args = arg_parser.parse_args(cli_args)

        if args.debug:
            _logger.setLevel("DEBUG")

        self.api_url = "https://api2.rhombussystems.com"
        self.device_id = args.device_id
        self.output = args.output
        self.use_wan = args.usewan

        if args.start_time:
            self.start_time = args.start_time
        else:
            now = datetime.now()
            self.start_time = int((now - timedelta(hours=1)).timestamp())
        if args.duration:
            self.duration = args.duration
        else:
            self.duration = 1 * 60 * 60  # 1 hour

        # initialize api http client
        self.api_sess = requests.session()

        # auth scheme changes depending on whether using cert/key or just api token
        if args.cert and args.private_key:
            scheme = "api"
            self.api_sess.cert = (args.cert, args.private_key)
        else:
            scheme = "api-token"

        self.api_sess.headers = {
            "x-auth-scheme": scheme,
            "x-auth-apikey": args.api_key}

        self.api_sess.verify = False

        self.media_sess = requests.session()
        self.media_sess.verify = False
        self.media_sess.headers = {
            "x-auth-scheme": scheme,
            "x-auth-apikey": args.api_key}

    def execute(self):
        # get a federated session token for media that lasts 1 hour
        session_req_payload = {"durationSec": 60 * 60}
        session_req_resp = self.api_sess.post(self.api_url + "/api/org/generateFederatedSessionToken",
                                              json=session_req_payload)
        _logger.debug("Federated session token response: %s", session_req_resp.content)

        if session_req_resp.status_code != 200:
            _logger.warn("Failed to retrieve federated session token, cannot continue: %s", session_req_resp.content)
            return

        federated_session_token = session_req_resp.json()["federatedSessionToken"]
        session_req_resp.close()

        # get camera media uris
        media_uri_payload = {"cameraUuid": self.device_id}
        media_uri_resp = self.api_sess.post(self.api_url + "/api/camera/getMediaUris",
                                            json=media_uri_payload)
        _logger.debug("Camera media uri response: %s", media_uri_resp.content)

        if session_req_resp.status_code != 200:
            _logger.warn("Failed to retrieve camera media uris, cannot continue: %s", media_uri_resp.content)
            return

        mpd_uri_template = media_uri_resp.json()["wanVodMpdUriTemplate"] if self.use_wan else \
            media_uri_resp.json()["lanVodMpdUrisTemplates"][0]

        _logger.debug("Raw mpd uri template: %s", mpd_uri_template)
        media_uri_resp.close()

        """ 
        When we make requests to the camera, the camera will use our session information to serve the correct files.
        The MPD document call starts the session and tells the camera the start time and duration of the clip requested
        We then get the seg_init.mp4 file which has the appropriate mp4 headers/init data
        and then we get the actual video segment files, named seg_1.m4v, seg_2.m4v, where each segment is a 2 second
        segment of video, so we need to go up to seg_<duration/2>.m4v.  The camera will automatically send the correct
        absolute time segments for each of the clip segments.  Concatenating the seg_init.mp4 and seg_#.m4v files into 
        a single .mp4 gives the playable video.
        """

        # the template has placeholders for where the clip start time and duration are supposed to go, so put the
        # desired start time and duration in the template
        mpd_uri = mpd_uri_template.replace("{START_TIME}", str(self.start_time)).replace("{DURATION}",
                                                                                         str(self.duration))
        _logger.debug("Mpd uri: %s", mpd_uri)

        # use the federated session token as our session id for the camera to process our requests
        media_headers = {"Cookie": "RSESSIONID=RFT:" + str(federated_session_token)}

        # start media session with camera by requesting the MPD file
        mpd_doc_resp = self.media_sess.get(mpd_uri, headers=media_headers)
        _logger.debug("Mpd doc: %s", mpd_doc_resp.content)
        mpd_info = RhombusMPDInfo(str(mpd_doc_resp.content, 'utf-8'))
        mpd_doc_resp.close()

        # start writing the video stream
        with open(self.output, "wb") as output_fp:
            # first write the init file
            init_seg_uri = get_segment_uri(mpd_uri, mpd_info.init_string)
            _logger.debug("Init segment uri: %s", init_seg_uri)

            init_seg_resp = self.media_sess.get(init_seg_uri, headers=media_headers)
            _logger.debug("seg_init_resp: %s", init_seg_resp)

            output_fp.write(init_seg_resp.content)
            output_fp.flush()
            init_seg_resp.close()

            # now write the actual video segment files.
            # Each segment is 2 seconds, so we have a total of duration / 2 segments to download
            for cur_seg in range(int(self.duration / 2)):
                seg_uri = get_segment_uri_index(mpd_info, mpd_uri,
                                                cur_seg)
                _logger.debug("Segment uri: %s", seg_uri)

                seg_resp = self.media_sess.get(seg_uri, headers=media_headers)
                _logger.debug("seg_resp: %s", seg_resp)

                output_fp.write(seg_resp.content)
                output_fp.flush()
                seg_resp.close()

                # log every 10 minutes of footage downloaded
                if cur_seg > 0 and cur_seg % 300 == 0:
                    _logger.info("Segments written from [%s] - [%s]",
                                 datetime.fromtimestamp(self.start_time + ((cur_seg - 300) * 2)).strftime('%c'),
                                 datetime.fromtimestamp(self.start_time + (cur_seg * 2)).strftime('%c'))
                    
            # finally rewrite the timestamps in the concatenated video to make it seekable
            self.adjust_timestamps()

        _logger.info("Succesfully downloaded video from [%s] - [%s] to %s",
                     datetime.fromtimestamp(self.start_time).strftime('%c'),
                     datetime.fromtimestamp(self.start_time + self.duration).strftime('%c'),
                     self.output)

    def adjust_timestamps(self):
        # check if ffmpeg is installed
        if os.system("which ffmpeg") == 0:
            # run the ffmpeg command
            adjusted_output = f"{self.output}_adjusted.mp4"
            command = f'ffmpeg -i "concat:{self.output}" -c copy {adjusted_output}'
            subprocess.run(command, shell=True)

            # replace the output file with the adjusted file
            os.remove(self.output)
            os.rename(adjusted_output, self.output)
        else:
            _logger.warning("ffmpeg not found on the system. Skipping timestamp adjustment.")        


    @staticmethod
    def __initialize_argument_parser():
        parser = argparse.ArgumentParser(
            description='Pulls footage from a Rhombus camera on LAN and stores it to the filesystem.')
        parser.add_argument('--api_key', '-a', type=str, required=True,
                            help='Rhombus API key')
        parser.add_argument('--device_id', '-d', type=str, required=True,
                            help='Device Id to pull footage from')
        parser.add_argument('--output', '-o', type=str, required=True,
                            help='The MP4 file to write to')
        parser.add_argument('--cert', '-c', type=str, required=False,
                            help='Path to API cert')
        parser.add_argument('--private_key', '-p', type=str, required=False,
                            help='Path to API private key')
        parser.add_argument('--start_time', '-s', type=int, required=False,
                            help='Start time in epoch seconds')
        parser.add_argument('--duration', '-u', type=int, required=False,
                            help='Duration in seconds')
        parser.add_argument('--debug', '-g', required=False, action='store_true',
                            help='Print debug logging')
        parser.add_argument('--usewan', '-w', required=False,
                            help='Use a WAN connection to download rather than a LAN connection', action='store_true')
        return parser


def main():
    # this cli command will save the last hour of footage from the specified device
    # python3 copy_footage_to_local_storage.py -a "<API TOKEN>" -d "<DEVICE ID>" -o out.mp4
    engine = CopyFootageToLocalStorage(sys.argv[1:])
    engine.execute()
    

if __name__ == "__main__":
    main()


import logging
from typing import Dict, Optional, Tuple, Union, cast

import aiofiles
import backoff
import httpx
from aiocache import cached
from backoff import _typing as backoff_typing
from httpx import HTTPError, HTTPStatusError, Timeout

from .lumeo_types import JsonObject

logging.getLogger("httpx").setLevel(logging.WARNING)
logging.getLogger("httpcore").setLevel(logging.WARNING)
logging.getLogger("aiocache").setLevel(logging.WARNING)

def backoff_handler(details: backoff_typing.Details) -> None:
    logging.warning(
        f"Backing off {details['wait']:0.1f} seconds after {details['tries']} tries"
        f" calling function {details['target'].__name__} with args {details['args']} and kwargs {details['kwargs']}"
    )

def backoff_no_giveup(exc: Exception) -> bool:
    return False

class LumeoApiClient:
    client: httpx.AsyncClient = httpx.AsyncClient(timeout=Timeout(60.0))

    def __init__(self, application_id: str, token: str, api_base_url: str = 'https://api.lumeo.com') -> None:
        self.headers: Dict[str, str] = {"Authorization": f"Bearer {token}"}
        self.base_url: str = api_base_url
        self.application_id = application_id
        self.client.base_url = self.base_url        
        self.client.headers = self.headers
        logging.info(f"Initialized Lumeo API client with base url: {self.base_url}, app_id: {application_id}, token: xxxx{token[-4:]}")

    def log_debug(self, message: str) -> None:
        logging.debug(f"{message}")
        
    def log_info(self, message: str) -> None:
        logging.info(f"{message}")

    def log_warning(self, message: str) -> None:
        logging.warning(f"{message}")

    def log_error(self, message: str) -> None:
        logging.error(f"{message}")
        
    async def close(self) -> None:
        await self.client.aclose()
        
    @backoff.on_exception(backoff.expo, HTTPError, max_time=60, on_backoff=backoff_handler, giveup=backoff_no_giveup, logger=None)        
    async def request(self, request_desc: str, method: str, url: str, timeout: Optional[Union[float, httpx.Timeout]] = None, **kwargs) -> Union[httpx.Response, None]:
        try:
            self.log_debug(f"API request: {request_desc} : {url}")
            response = await self.client.request(method, url, headers=self.headers, timeout=timeout, **kwargs)
            response.raise_for_status()
            return response
        except HTTPError as err:            
            # Log error    
            print(f"{err} : {err.response.text}")
            message = err.response.status_code if isinstance(err, HTTPStatusError) else err #error.response.text
            self.log_warning(f"API request failed. Will retry. {request_desc}: {message}")
            
            # If the error is a 400, 401, 403, 404, 409, dont raise exception, return None
            if isinstance(err, HTTPStatusError) and err.response.status_code in {400, 401, 403, 404, 409}:
                return None                    
            
            # Re-raise the exception to be caught by backoff
            raise
        except Exception as e:
            # This will catch the HTTPError re-raised by backoff after max_time is reached
            self.log_error(f"API Error {request_desc}: {message}")
            return None
        
    async def create_event(
        self, event_type: str, severity: str, payload: str, 
        deployment_id: Union[str, None], camera_id: Union[str, None], stream_id: Union[str, None]
    ) -> None:
        event_json = {
            "category": "ftp-gateway",
            "event_type": event_type,
            "severity": severity,
            "payload": payload,
            "context": None,
            "related_entities": {
                "deployment_id": None,
                "camera_id": None,
                "stream_id": None,
                "gateway_id": None,
                "file_id": None,
                "node_id": None,
            },
        }

        if deployment_id:
            event_json["object"] = "deployment"
            event_json["object_id"] = deployment_id
            event_json["related_entities"]["deployment_id"] = deployment_id

        if camera_id:
            event_json["object"] = "camera"
            event_json["object_id"] = camera_id
            event_json["related_entities"]["camera_id"] = camera_id
            
        if stream_id:
            event_json["object"] = "stream"
            event_json["object_id"] = stream_id
            event_json["related_entities"]["stream_id"] = stream_id

        await self.request(f"Creating event {event_type}", "POST", f"/v1/apps/{self.application_id}/events", json=event_json)

        return

    async def get_camera_with_external_id(self, external_id: str) -> Union[JsonObject, None]:
        response = await self.request(
            f"Getting camera with external id {external_id}",
            "GET",
            f"/v1/apps/{self.application_id}/cameras",
            params={"pagination": "cursor", "limit": 1, "external_ids[]": external_id},
        )
        response_json = response.json()
        if "data" in response_json and len(response_json["data"]) > 0:
            return response_json["data"][0]
        else:
            return None
        
    async def get_camera_with_id(self, camera_id: str) -> Union[JsonObject, None]:
        response = await self.request(
            f"Getting camera with id {camera_id}",
            "GET",
            f"/v1/apps/{self.application_id}/cameras/{camera_id}"
        )        
        if response:            
            return response.json()
        else:
            return None        

    async def create_virtual_camera(self, external_id: str, name: str) -> JsonObject:
        response = await self.request(
            f"Creating virtual camera with external id {external_id}",
            "POST",
            f"/v1/apps/{self.application_id}/cameras",
            json={
                "external_id": external_id,
                "name": name,
                "model": "Virtual",
                "status": "unknown",
                "uri": None,
                "conn_type": "virtual",
            },
        )
        response_json = response.json()
        return response_json

    async def set_camera_reference_deployments(self, camera_id: str, deployment_id: str) -> None:
        await self.request(
            f"Setting camera reference deployment for camera {camera_id}",
            "POST",
            f"/v1/apps/{self.application_id}/cameras/{camera_id}/reference_deployments",
            json=[deployment_id],
        )

    @cached(ttl=3600)
    async def get_deployment_queue_id(self) -> str:
        response = await self.request(f"Getting deployment queue", "GET", f"/v1/apps/{self.application_id}/deployment_queues")
        response_json = response.json()
        return response_json and response_json[0]["id"]

    async def create_file(self, file_name: str, file_size: int, camera_id: Union[str,None]) -> JsonObject:
        response = await self.request(
            f"Creating file with file name {file_name}",
            "POST",
            f"/v1/apps/{self.application_id}/files",
            json={
                "name": file_name,
                "size": file_size,
                "gateway_id": None,
                "pipeline_id": None,
                "node_id": None,
                "deployment_id": None,
                "camera_id": camera_id,
                "stream_id": None,
                "metadata": None,
                "description": None,
            },
        )
        response_json = response.json()
        return response_json

    async def upload_file(self, data_url: str, metadata_url: str, file_path: str) -> None:
        self.log_debug(f"API Uploading file {file_path} ...")
        try:
            async with aiofiles.open(file_path, "rb") as f:
                file_content = await f.read()
                response = await self.client.put(data_url, data=file_content)
                response.raise_for_status()

            await self.client.put(metadata_url, data="null", timeout=3600.0)
            return True
        except HTTPError as error:
            message = error.response.text if isinstance(error, HTTPStatusError) else error
            self.log_error(f"Error while uploading file {file_path}: {message}")

        return False

    async def set_file_status(self, file_id: str, status: str) -> None:
        await self.request(
            f"Setting file status for file {file_id}",
            "PUT",
            f"/v1/apps/{self.application_id}/files/{file_id}/cloud_status",
            data=status,
        )
        
    async def add_tag_to_file(self, file_id: str, tag_id: str) -> None:
        await self.request(
            f"Adding tag {tag_id} to file {file_id}",
            "POST",
            f"/v1/apps/{self.application_id}/files/{file_id}/tags",
            json=[tag_id],
        )

    async def create_lumeo_file_stream(self, file: JsonObject, camera_id: Union[str,None]) -> JsonObject:
        return await self.create_file_stream(file['name'], f"lumeo://{file['id']}", camera_id)
    
    async def create_file_stream(self, name: str, url: str, camera_id: Union[str,None]) -> JsonObject:        
        response = await self.request(
            f"Creating file stream for file {name}",
            "POST",
            f"/v1/apps/{self.application_id}/streams",
            json={
                "name": name[:200],
                "uri": url,
                "source": "uri_stream",
                "stream_type": "file",
                "camera_id": camera_id,
                "gateway_id": None,
                "status": "unknown",
            },
        )
        response_json = response.json()
        return response_json
    
    async def create_live_stream(self, name: str, url: str, gateway_id: Union[str,None]=None) -> JsonObject:        
        response = await self.request(
            f"Creating live stream for file {name}",
            "POST",
            f"/v1/apps/{self.application_id}/streams",
            json={
                "name": name[:200],
                "uri": url,
                "source": "uri_stream",
                "stream_type": "rtsp",
                "camera_id": None,
                "gateway_id": gateway_id,
                "status": "unknown",
            },
        )
        response_json = response.json()
        return response_json
    
    async def get_streams(self, tag_id: str, stream_type: str="file") -> JsonObject:
        all_streams = []
        page = 1
        while True:
            response = await self.request(
                f"Getting file streams for tag {tag_id} (page {page})",
                "GET",
                f"/v1/apps/{self.application_id}/streams",
                params={"pagination": "offset", "limit": 50, "page": page, "stream_types[]": stream_type, "tagged_with[]": tag_id,
                        "include_tagged_with_descendants": False, "only_untagged": False}
            )
            
            response_json = response.json()            
            streams = response_json.get('data', [])
            all_streams.extend(streams)
            
            if len(streams) < 50:  # Less than the limit, so it's the last page
                break
            else:            
                page += 1        
        return all_streams
        
    @cached(ttl=3600)
    async def get_stream_with_id(self, stream_id: str) -> Union[JsonObject, None]:
        response = await self.request(
            f"Getting stream with id {stream_id}",
            "GET",
            f"/v1/apps/{self.application_id}/streams/{stream_id}"
        )        
        if response:            
            return response.json()
        else:
            return None     
        
    async def add_tag_to_stream(self, stream_id: str, tag_id: str) -> None:
        await self.request(
            f"Adding tag {tag_id} to stream {stream_id}",
            "POST",
            f"/v1/apps/{self.application_id}/streams/{stream_id}/tags",
            json=[tag_id],
        )  
    
    async def get_tag_id_by_path(self, tag_path: str) -> JsonObject:
        tag_names = tag_path.split("/")
        current_tag_id = None
        for tag_name in tag_names:
            params = {"pagination": "offset", "page": 1, "limit": 1, "tag_names[]": tag_name}
            if current_tag_id:
                params['parents[]'] = current_tag_id
                
            response = await self.request(
                f"Getting tag {tag_name}", "GET", f"/v1/apps/{self.application_id}/tags", params=params
            )
            if response:
                tags_list = response.json()
                if tags_list['data']:
                    current_tag_id = tags_list['data'][0]['id']
                else:
                    return None
            else:
                return None
        self.log_debug(f"Tag leaf {tag_path} found with id {current_tag_id}")                
        return current_tag_id
    
    async def create_tag_path(self, tag_path: str) -> JsonObject:
        tag_names = tag_path.split("/")
        current_tag_id = None
        for tag_name in tag_names:
            # Check if the tag already exists
            existing_tag = await self.get_tag_id_by_path("/".join(tag_names[:tag_names.index(tag_name)+1]))
            if existing_tag:
                current_tag_id = existing_tag
            else:
                # Create the tag if it doesn't exist
                response = await self.request(
                    f"Creating tag {tag_name}", "POST", f"/v1/apps/{self.application_id}/tags",
                    json={
                        "name": tag_name,
                        "parent": current_tag_id,
                    },
                )
                if response:
                    response_json = response.json()
                    current_tag_id = response_json["id"]
                else:
                    return None
        return current_tag_id
    
    @cached(ttl=3600)
    async def get_pipeline(self, pipeline_id: str) -> Union[JsonObject, None]:
        response = await self.request(
            f"Getting pipeline {pipeline_id}", "GET", f"/v1/apps/{self.application_id}/pipelines/{pipeline_id}"
        )
        response_json = response.json()
        return response_json

    @cached(ttl=3600)
    async def get_deployment(self, deployment_id: str) -> Union[JsonObject, None]:
        response = await self.request(
            f"Getting deployment {deployment_id}", "GET", f"/v1/apps/{self.application_id}/deployments/{deployment_id}"
        )
        response_json = response.json()
        return response_json

    async def queue_deployment(
        self, queue_id: str, pipeline_id: str, deployment_configuration: JsonObject, deployment_name: Union[str, None]
    ) -> JsonObject:
        response = await self.request(
            f"Queueing deployment for pipeline {pipeline_id} with name '{deployment_name}', configuration {deployment_configuration}",
            "POST",
            f"/v1/apps/{self.application_id}/deployment_queues/{queue_id}/entries",
            json={
                "pipeline_id": pipeline_id,
                "deployment_name": deployment_name,
                "deployment_configuration": deployment_configuration,
            },
        )
        response_json = response.json()
        return response_json
    
    async def create_deployment(
        self, pipeline_id: str, gateway_id: str, deployment_configuration: JsonObject, deployment_name: Union[str, None]
    ) -> JsonObject:
        response = await self.request(
            f"Creating deployment for pipeline {pipeline_id} with name '{deployment_name}', configuration {deployment_configuration}",
            "POST",
            f"/v1/apps/{self.application_id}/deployments",
            json={
                "pipeline_id": pipeline_id,
                "gateway_id": gateway_id,
                "name": deployment_name,
                "state": "running",
                "configuration": deployment_configuration,
            },
        )
        response_json = response.json()
        return response_json

    async def get_queue_entries(self, queue_id: str) -> JsonObject:
        response = await self.request(
            f"Getting queue info for queue {queue_id}", 
            "GET", 
            f"/v1/apps/{self.application_id}/deployment_queues/{queue_id}/entries",
            params={"pagination": "offset", "page": 1, "limit": 1},            
        )
        response_json = response.json()
        return response_json



import pkgutil

def list():
    """
    Lists all the executable scripts within all subpackages and modules of this package.

    Returns:
        list: A list of strings representing the paths to the executable scripts.
    """
    scripts = []
    package = __import__('lumeo.scripts', fromlist=[''])
    for importer, modname, ispkg in pkgutil.walk_packages(package.__path__, package.__name__ + '.'):
        if not ispkg:
            module = __import__(modname, fromlist=[''])
            if hasattr(module, 'main'):
                scripts.append(modname)

    sanitized_scripts = []    
    for script in scripts:
        sanitized_script = script.replace('.scripts.', '.').replace('.__main__', '').replace('_', '-').replace('.','-')  
        sanitized_scripts.append(sanitized_script)
        print(sanitized_script)
                      
    return True

if __name__ == "__main__":
    list()


import unittest
import numpy as np
import cv2
from lumeo.pipeline.display import write_label_on_frame

class TestDisplay(unittest.TestCase):
    def test_write_label_on_frame(self):
        # Create a black image
        mat = np.zeros((500, 500, 3), dtype=np.uint8)

        # Define label and position
        xidx, yidx = 50, 50
        label = "This is a test label. It should wrap correctly."

        # Call the method
        label_width, total_label_height = write_label_on_frame(mat, xidx, yidx, label)

        # Check that label is written on the frame by verifying pixel values
        self.assertNotEqual(np.sum(mat), 0, "The label was not written on the frame.")

        # Optionally, you can check specific pixels to verify that the label is correctly placed
        self.assertTrue(np.any(mat[50:50+total_label_height, 50:50+label_width] != 0), "Label not correctly written.")

if __name__ == '__main__':
    unittest.main()
